/******************************************************************************\
 * DISKACCT.C -- Summarize disk usage by UIC                                  *
\******************************************************************************/
#include descrip.h
#include libdef.h
#include file.h
#include stdio
#include "ods2.h"
#line 9 "diskacct.c"

struct NODE {
        struct  NODE    *node$a_left;
        struct  NODE    *node$a_right;
        short           node$w_mbz;
        unsigned long   node$l_uic;
        long            node$l_cnt;
} *root = 0;

long compare(uic, node, cnt)
struct NODE *node;
unsigned long *uic;
long *cnt;
{       if (*uic == node->node$l_uic)
                node->node$l_cnt  += *cnt;
        return (*uic > node->node$l_uic ? 1 : *uic < node->node$l_uic ? -1 : 0);
}

long allocate (uic, newnode, cnt)
struct NODE **newnode;
unsigned long *uic;
long *cnt;
{       struct NODE *node;
        long i;

        i = sizeof ( struct NODE );
        if (((i = LIB$GET_VM(&i, newnode)) && 7) != 1)
                exit(i);
        node = *newnode;
        node->node$a_left = 0;
        node->node$a_right = 0;
        node->node$w_mbz = 0;
        node->node$l_uic = *uic;
        node->node$l_cnt = *cnt;
        return(1);
}

long action(node, n)
struct NODE *node;
long n;
{       printf("[%06o,%06o]         %d\n",(node->node$l_uic & 0xFFFF0000) >> 16,
                node->node$l_uic & 0x0000FFFF, node->node$l_cnt);
        return(1);
}

struct FILE_HEADER header;
struct HOME_BLOCK home;
#define FIRST_FID_INDEX (4 * home.SBCL + home.IBSZ)

main(nargs, args)
int nargs;
char **args;
{   long fid;
    unsigned long map_bytes, i, count, header_num, level, status;
    unsigned char format;
    unsigned short cksm, *ptr;
    unsigned long FFFFFFFF = 0xFFFFFFFF, uic;
    unsigned short FFFF = 0xFFFF;
    struct NODE *newnode;
    long ctlflag = 0;
    struct RTR_PTR_00 *rp_0;
    struct RTR_PTR_01 *rp_1;
    struct RTR_PTR_10 *rp_2;
    struct RTR_PTR_11 *rp_3;

    if (nargs < 2)
    {   fprintf(stderr, "Usage:  diskacct index_file_name\n");
        exit(0);
    }
    ioinit(&nargs, args);
    if ((fid = open(*++args, O_RDONLY, 0, "shr=put", "mbc=100")) < 0)
        exit(0);
    if (read(fid, (char *) &home, 512) != 512)
        exit(0);
    if (read(fid, (char *) &home, 512) != 512)
        exit(0);
    for (i = FIRST_FID_INDEX - 2; i > 0; --i)
        if (read(fid, (char *) &header, 512) != 512)
            exit(0);
    header_num = 0;
    while (read(fid, (char *) &header, 512) == 512)
    {   for (ptr = (char *) &header, cksm = 0, i = 0; i < 255; ++i, ++ptr)
            cksm += *ptr;
        header_num++;
        if ((cksm == header.CKSM || (header.CKSM == 0125252 &&
             header.H.SCHA & 16 == 0)) &&
            (header.H.IDOF >= (short *) &header.H.FOWN - (short *) &header) &&
            (header.H.IDOF <= header.H.MPOF) &&
            (header.H.MPOF <= header.H.ACOF) &&
            (header.H.ACOF <= header.H.RSOF) &&
            ((header.H.FLEV & 0xFF00) == 0x200) &&
            ((header.H.FLEV & 0xFF) > 0) &&
            (header.H.FNUM == header_num) &&
            (header.H.USE <= header.H.ACOF - header.H.MPOF))
        {   rp_0 = (struct RTR_PTR_00 *) (((unsigned short *) &header) +
 header.H.MPOF);
            uic = header.H.FOWN;
            for (map_bytes =  2 * header.H.USE, count = 1; map_bytes > 0;)
            {   switch (format = rp_0->FORMAT)
                {   case 0:
                        rp_0++;
                        map_bytes -= sizeof(struct RTR_PTR_00);
                        break;
                    case 1:
                        rp_1 = (struct RTR_PTR_01 *) rp_0;
                        if (rp_1->HI_LBN_01 != 0x03F ||
                            rp_1->LO_LBN_01 != FFFF)
                            count += rp_1->COUNT_01 + 1;
                        map_bytes -= sizeof(struct RTR_PTR_01);
                        rp_1++;
                        rp_0 = (struct RTR_PTR_00 *) rp_1;
                        break;
                    case 2:
                        rp_2 = (struct RTR_PTR_10 *) rp_0;
                        if (rp_2->LBN_10 != FFFFFFFF)
                            count += rp_2->COUNT_10 + 1;
                        map_bytes -= sizeof(struct RTR_PTR_10);
                        rp_2++;
                        rp_0 = (struct RTR_PTR_00 *) rp_2;
                        break;
                    case 3:
                        rp_3 = (struct RTR_PTR_11 *) rp_0;
                        if (rp_3->LBN_11 != FFFFFFFF)
                            count += (rp_3->HI_CNT_11 << 16) +
                                rp_3->LO_CNT_11 + 1;
                        map_bytes -= sizeof(struct RTR_PTR_11);
                        rp_3++;
                        rp_0 = (struct RTR_PTR_00 *) rp_3;
                        break;
                    default:
                        puts("Invalid retrieval pointer encountered\n");
                        exit(0);
                }
            } if (((status = LIB$INSERT_TREE(&root, &uic, &ctlflag, compare,
                allocate, &newnode, &count)) & 7) != 1 &&
                status != LIB$_KEYALRINS)
                exit(status);
        }
    } exit(LIB$TRAVERSE_TREE(&root, action, 0));
}

/* IOINIT.C -- I/O redirection subroutine.  Redirects stdin (< or <<);
 *      stdout (> or >>); or stderr ( 2>, 2>>)
 * Usage:
 *      main(nargs, args)
 *      {       ...
 *              ioinit(&nargs, args);
 *              ...
 *      }
 */

#include <errno.h>
#include <stdio.h>

typedef char *STR;

ioinit(nargs, args)
int *nargs;
STR *args;
{   int argnum;

    for (argnum = 1; argnum < *nargs; ++argnum)
    {   if (strncmp(args[argnum], "<", 1) == 0)
        {   args[argnum] += 1;
            *nargs = u_reopen(*nargs, args, argnum, "r", stdin);
            argnum--;
        } else if (strncmp(args[argnum], ">>", 2) == 0)
        {   args[argnum] += 2;
            *nargs = u_reopen(*nargs, args, argnum, "a", stdout);
            argnum--;
        } else if (strncmp(args[argnum], ">", 1) == 0)
        {   args[argnum] += 1;
            *nargs = u_reopen(*nargs, args, argnum, "w", stdout);
            argnum--;
        } else if (strncmp(args[argnum], "2>>", 3) == 0)
        {   args[argnum] += 3;
            *nargs = u_reopen(*nargs, args, argnum, "a", stderr);
            argnum--;
        } else if (strncmp(args[argnum], "2>", 2) == 0)
        {    args[argnum] += 2;
            *nargs = u_reopen(*nargs, args, argnum, "w", stderr);
            argnum--;
        }
    }
}

u_reopen(nargs, args, argnum, acmod, chan)
int nargs, argnum;
STR *args, acmod;
FILE *chan;
{   char *file;
    int offset, i, errornum;

    if (args[argnum][0] != '\0')
    {   offset = 1;
        file = args[argnum];
    } else if ((argnum + 1) < nargs)
    {   offset = 2;
        file = args[argnum+1];
    } else
    {   fprintf(stderr, "Illegal redirection on command line.\n");
        exit(1);
    } for (i = argnum; i < nargs - offset; ++i)
        args[i] = args[i + offset];
    if (((*acmod == ' ' || strcmp(acmod,"a+") == 0) && freopen(file, acmod,
        chan ,"rfm=stm") != chan) || freopen(file, acmod, chan) != chan)
    {   if (errno == EVMSERR)
        {       errornum = vaxc$errno;
                fprintf(stderr, "Failure opening redirected stream.\n");
                exit(errornum);
        } else
        {       perror("Failure opening redirected stream.");
                exit(1);
        }
    } return(nargs - offset);
}

